@@ -20,7 +20,7 @@ get.current.packages <- function(type) {
         } else {
   #Else get available packages to see if each attempted to install is new
       current.packages <- tryCatch({
-          as.data.frame(utils::available.packages(type = type)[, c(1, 2)],stringsAsFactors = FALSE)
+          data.frame(utils::available.packages(type = type)[, c(1, 2)],stringsAsFactors = FALSE,row.names=NULL)
         },
         error = function(e) NULL)
 		
