@@ -1,7 +1,6 @@
 #This script has functions used throughout the package
-
 #1  get.pkg(), get.vsr()     : Extract package and version information from pkg_vrs
-#2  is.pkg_vrs.installed()   : Is pkg_vrs installed (within same R-minor version)?
+#2  is.pkg_vrs.installed()   : Is pkg_vrs installed (within same R-minor version)? 
 #3  as.DateYMD               : Format Y-M-D as date
 #4  get.rdate()              : Date for R being used
 #5  get.rversion()           : Get R Version major:minor:patch
@@ -10,7 +9,7 @@
 #8  namedList()              : automatically name object in list 
 #9  quit.menu()              : prompt to quit call upon mismatch oR0f dates
 #10 exit()                   : Stop message which does not say error
-#11 get.available.mran.date: Available mran dates
+#11 Dropped (mran related)
 #12 base_pkg()                : Output vector with all base packages
 #13 ignore.deps_default()     :  Default packages to ignore conflicts with (but gives warning)
 #14 Is this on R studio
@@ -19,7 +18,7 @@
 #17 validate.date()           : Validate date  
 #18 strip.prompt)_            : Clean up prompt answer, lowercase and no quotes
 #19 exit.if.download.failed() : Verify file was download
-#20 Check if lib is for github (possibly a legacy function)
+#20 Check if lib is for GitHub (possibly a legacy function)
 #21 Robust reading of text online : try_readlines ()
 #22 Turn dates to unix time
 #23 strpos1()                - string position - simpler gregexpr for strpos  
@@ -34,21 +33,40 @@
 #32 infinite.prompt()          :  ask the same question until a valid answer is provided
 #33 get.packages_df()          :  data.frame with installed packages in local library
 #34 sandwich.library()         :  turn a string containing library calls into a vector of pkg names
-#35 format.msg()               :  format output to have fixed width and starting symbol (e.g., "|    ")
+#35 format_msg()               :  format output to have fixed width and starting symbol (e.g., "|    ")
 #36 set.default.mirror()       :  set a CRAN mirror if none is already set
 #37 verify.personal.library.exists(): if no folder for saving personal packages exists, prompted to create it
-#38 restart.text()             :  tells user to either quit() R or use CMD-SHIF-F10 based on whether they use R Studio
+#38 restart.text()             :  tells user to either quit() R or use CMD-SHFT-F10 based on whether they use R Studio
 #39 get.minutes.since.cookie() :  reads csv file with Sys.time() of last time this cookie was saved
-#40 get.repos_from.snowball()  :  infer repository (cran, github, gitlab) from other information already in the snowball
-#41 update.groundhog.session() :  Update groundhog session with new packages loaded and requested with groundhog
+#40 get.repos_from.snowball()  :  infer repository (cran, GitHub, GitLab) from other information already in the snowball
+#41 add.session.snowballs() :     Update session dataframe with info on all groundhog loaded snowballs
 #42 check.groundhog.version()  :  if more than `min.days' days since last check, check if groundhog needs to be updated
-########################################################################
+#43 Get operating system       :  windows or mac or mac_arm?
+#44 Check consent              :  Does .../R_groundhog exists?
+#45 Turn path from '\' to '/'  :  c:\users --> c:/users
+#46 MOVED
+#47 DROPPED
+#48 message.batch.installation.feedback() : feedback while installing a batch in parallel
+#49 - get.parallel.time()  :  estimate parallel time of installation
+#50 get pkg_list from path
+#51 gstop()                    :  show message in green then "** groundhog stopped**" in red and exit())
+#52 read/save loadl rds 
+#53 Download in batches        : breaks a list of pkgs into batches downloaded sequentially
+#54 filesize_format            : turn bytes file size to human readable
+#55 get.restore.points         : vector with dates available 
+#57 View conflicts             : view conflicting pkgs in recent call
+#58 Get ip.groundhog():        : installed.packages data.frame for all packages in groundhog library and loans
+#59 get ip.backup()            : same for all packages that have been saved to the backup folder
+#60 get.loans() & /save.loans(): load and save database with all lent packages
+#61 file.rename.robust:        :  rename a file ony once it can be renamed to itself
+
+####################################################################################
     
 
 
-#1. get.pkg get.vrs -  Exract package and version information from pkg_vrs
-  get.pkg <- function(x) substr(x, 1, regexpr("_", basename(x)) - 1)
-  get.vrs <- function(x) substr(x, regexpr("_", basename(x)) + 1, nchar(x))
+#1. get.pkg get.vrs -  Extract package and version information from pkg_vrs
+  get.pkg <- function(x) substr(x, 1, regexpr("_", basename(as.character(x)))- 1)
+  get.vrs <- function(x) substr(x, regexpr("_", basename(as.character(x))) + 1, nchar(x))
 
 #2.  Is pkg_vrs installed (within same R-minor version)?
       is.pkg_vrs.installed <- function(pkg, vrs) {
@@ -65,7 +83,7 @@
 #3.  Format Y-M-D as date
   as.DateYMD <- function(x) as.Date(x, format = "%Y-%m-%d",origin='1970-01-01')
 
-#4. R beind used
+#4. R being used
   get.rdate <- function() {
     date <- paste0(R.version$year, "-", R.version$month, "-", R.version$day)
     return(as.DateYMD(date))
@@ -144,10 +162,10 @@
       message1("You typed ", x, " so script stops...")
       msg.R.switch(date)
       exit("---")
-    } # End if quit
+    } # end if quit
   
     message1("You typed '", x, "' the script continues...")
-  } # End quit.menu
+  } # end quit.menu
 
 #10 exit() Stop message which does not say error
   exit <- function(...) {
@@ -155,43 +173,18 @@
     
      #Read cran toc again to undo any changes with remote
         #Only run if main folder exists, otherwise we have loop, 
-        #exiting when it does not exist, and asking that it exist when exitin
+        #exiting when it does not exist, and asking that it exist when exiting
           main_folder <-  paste0(path.expand("~"), "/R_groundhog")
           if (file.exists(main_folder)) {
             .pkgenv[['cran.toc']] <- readRDS(file.path(get.groundhog.folder(),"cran.toc.rds"))
           }     
           
-    #Return libpath
-       .libPaths(.pkgenv[["orig_lib_paths"]])
+    #Return .libpath if it has been changed
+       if (!is.null(.pkgenv[["orig_lib_paths"]])) .libPaths(.pkgenv[["orig_lib_paths"]])
         invokeRestart("abort")
   }
 
-#11 Available mran dates
-get.available.mran.date <- function(date0, date1) {
-  missing.mran.dates <- .pkgenv[["missing.mran.dates"]]
 
-  all.dates <- date0:date1 # All dates in range
-  available.dates <- all.dates[!all.dates %in% missing.mran.dates] # Those that are not missing
-  if (length(available.dates) == 0) {
-    return(as.Date("1970-01-01"))
-  } # If none remain, end
-
-  # Report mid value
-  n.dates <- length(available.dates)
-
-  if (n.dates == 0) {
-    message1(
-      "We looked for the version of the package you need in MRAN ",
-      "but it was not found there"
-    )
-    exit()
-  }
-
-  # ceiling() rather than floor() or round() to work when n.dates <- 1
-  mid.date.k <- ceiling(n.dates / 2)
-  mid.date <- available.dates[mid.date.k]
-  return(as.Date(mid.date, origin = "1970-01-01"))
-} # End of function
 
 #12 Base packages
     base_pkg <- function() {
@@ -213,8 +206,8 @@ get.available.mran.date <- function(date0, date1) {
         )
   
   #Recommended and thus hard to uninstall packages
-      ip <- data.frame(utils::installed.packages(),stringsAsFactors = FALSE)
-      recommended.pkgs <- unique(subset(ip, ip$Priority=="recommended")$Package) #unique becuase there may be two versions of the same package in different libraries
+      ip <- data.frame(utils::installed.packages(),stringsAsFactors = FALSE,row.names=NULL)
+      recommended.pkgs <- unique(subset(ip, ip$Priority=="recommended")$Package) #unique because there may be two versions of the same package in different libraries
       
   #Combine
       ignore.deps <- c(Rstudio.deps, recommended.pkgs)
@@ -242,7 +235,7 @@ get.available.mran.date <- function(date0, date1) {
      return(majmin)
      }
    
-#16 Get major minor pach
+#16 Get major minor patch
  get.r.majmin.release <- function()
  {
    r.majmin <- get.r.majmin()
@@ -256,34 +249,34 @@ get.available.mran.date <- function(date0, date1) {
 #17 validate date  
   validate.date <- function(entered_date)
       {
-       msg=paste0("\ngroundhog says: The date you entered '", entered_date,"', is not valid.\n",
+       msg <- paste0("The date you entered '", entered_date,"', is not valid.\n",
                 "Please use the 'yyyy-mm-dd' format"
           )
     
+       #more than one
+       if (length(entered_date)>1) gstop("The 'date' argument in groundhog.library() must have only one value.")
+       
        
        # correct format
         d <- try(as.Date(entered_date, format="%Y-%m-%d"),silent = TRUE)
           if ("try-error" %in% class(d) || is.na(d)) {
-              message1(msg)
-               exit()
+             gstop(msg) #util #51)
           }
 
        
               
-      #The format check does not verify that the day is at most 2 charcters long,  (e.g., it accepts 2022-01-109)
+      #The format check does not verify that the day is at most 2 characters long,  (e.g., it accepts 2022-01-109)
        if (is.character(entered_date)) { 
        d.parts <- strsplit(entered_date,"-")[[1]]     #split date by '-
         if (nchar(d.parts[3])>2) {
-            message1(msg)
-            exit()
+           gstop(msg) #util #51)
         }
        }
    
        
       # numeric
          if (is.numeric(entered_date)) {
-          message1(msg)
-          exit()
+           gstop(msg) #util #51)
         }
  
            
@@ -301,7 +294,7 @@ get.available.mran.date <- function(date0, date1) {
   }
   
   
-#19 Verify file was download
+#19 Verify file was downloaded
     exit.if.download.failed <- function(file.name,file.path)
     {
     if (!file.exists(file.path)) {
@@ -316,9 +309,9 @@ get.available.mran.date <- function(date0, date1) {
     }#End if path exists
   } #End function
 
-#20 Check if lib is for github
-    is.lib.github=function(lib.path) {
-      pos = regexpr('_github', lib.path)[[1]] 
+#20 Check if lib is for GitHub
+    is.lib.GitHub=function(lib.path) {
+      pos = regexpr('_GitHub', lib.path)[[1]] 
       return(pos>0) #TRUE if pos>0
       
       }
@@ -357,7 +350,7 @@ get.available.mran.date <- function(date0, date1) {
       date.to.time      <- function(date)       as.numeric(as.POSIXct(date,      format="%Y-%m-%d"))
       
 
-#Function 23 - string positiong - simpler gregexpr for strpos
+#Function 23 - string position - simpler gregexpr for strpos
       strpos1 <- function(needle, haystack) as.numeric(gregexpr(needle, haystack)[[1]])
       
 
@@ -416,8 +409,8 @@ get.available.mran.date <- function(date0, date1) {
             if (x==TRUE | x==FALSE) {
               
                return() } else {
-                message('groundhog says: "' , substitute(x) , '" can only be TRUE or FALSE')
-                 exit()
+                msg=paste0('groundhog says: "' , substitute(x) , '" can only be TRUE or FALSE')
+                 gstop(msg) #util #51)
                }
         }
       
@@ -468,7 +461,8 @@ get.available.mran.date <- function(date0, date1) {
     
     infinite.prompt <- function(text_msg, valid_answers,must.restart=FALSE)
       {
-  
+      
+        
       
       #Initialize values
         answer <- ''
@@ -484,7 +478,7 @@ get.available.mran.date <- function(date0, date1) {
           text.max<-20
           answer2 <- ifelse(nchar(answer)>text.max, paste0(substr(answer,1, text.max),"..."), answer)
       
-        #Draft text showign answer
+        #Draft text showing answer
           msg.answer <- paste0('You typed --> "' , answer2 , '".  That is sadly not an accepted response.')
         
         #Add counter if k>1
@@ -505,7 +499,7 @@ get.available.mran.date <- function(date0, date1) {
         
         #Add counter
           k=k+1
-        } #end while
+        } #End while
         return(answer)
     } #End prompt    
 
@@ -549,13 +543,13 @@ get.available.mran.date <- function(date0, date1) {
           x <- strsplit(x," ")[[1]]
           x <- x[x!=""]
           return(x)
-        } #ENd of function
+        } #End of function
     
     
 #35 Format msg: format output to have fixed width and starting symbol (e.g., "|    ")
     
     
- format.msg <- function(msg,width=70, header='IMPORTANT.', pre="|")
+ format_msg <- function(msg,width=70, header='IMPORTANT.', pre="|   ")
 {
   #Line counter
     j<-0
@@ -618,27 +612,50 @@ get.available.mran.date <- function(date0, date1) {
   
  }
  
-#37 Does personal folder to install R packages exist?
+#37 Does personal folder to install R packages (not groundhog, but R's default) exist
     verify.personal.library.exists<-function()
       {
       #Create personal library if it does not exist
-        default_library <- Sys.getenv("R_LIBS_USER")
+         default_library <- Sys.getenv("R_LIBS_USER")
       
-      if (length(.libPaths()) <= 1 & !file.exists(default_library)) {
-        msg <- paste0("R does not have a personal library to save packages to. ",
-               "The default location for it is: '", default_library,"'. \n ",
-               "   1) Type 'create' to create that directory \n ",
-               "   2) Otherwise type 'stop'")
-        
-        answer<-infinite.prompt(format.msg(msg),c('create','stop'))
-        if (answer=='create') {
-              dir.create(Sys.getenv("R_LIBS_USER"), recursive = TRUE)
-              }
-        
-        if (answer=='stop') {
-              exit()
-        }
+         
+      #Interactive Session
+        if (length(.libPaths()) <= 1 & !file.exists(default_library))
+          {
+          
+          #Start msg
+              msg <- paste0("R does not have a personal library to save packages to. ",
+                            "The default location for it is: '", default_library,"'. \n ")
+          
+          
+          #Interactive menu
+            if (interactive()==TRUE) {
+                  msg <-paste0(msg,
+                       "   1) Type 'create' to create that directory \n ",
+                       "   2) Otherwise type 'stop'")
+                
+                answer<-infinite.prompt(format_msg(msg),c('create','stop'))
+                if (answer=='create') {
+                      dir.create(Sys.getenv("R_LIBS_USER"), recursive = TRUE)
+                      }
+                
+                if (answer=='stop') {
+                      exit()
+                }} 
+          #If script
+            if (interactive()==FALSE) {
+                  msg <- paste0("To work with groundhog you need to create that library.",
+                                 "One way to do that is to run `dir.create('", default_library , "',recursive=TRUE)`")
+
+                  message(msg)
+                  exit()
+                }
+                
+                
       }#End if library does not exist
+         
+       
+    
     }#End function
         
 #38 restart.text()  - give instructions for restarting in R
@@ -655,7 +672,6 @@ get.available.mran.date <- function(date0, date1) {
       
     }
 
-    cookie_name='warng1'
     
 #39 save.cookie() and read.cookie()
     
@@ -673,81 +689,117 @@ get.available.mran.date <- function(date0, date1) {
           utils::write.csv(as.numeric(Sys.time()),cookie_path,row.names = FALSE)
         }
         
-   #39.2 READ
+     #39.2 Does cookie exist
+        cookie.exists<-function(cookie_name)
+        {
+          cookies_dir <- paste0(get.groundhog.folder(),"/cookies")
+          cookie_path <- file.path(cookies_dir, paste0(cookie_name,".csv"))
+          return (file.exists(cookie_path)) 
+       }
+          
+        
+   #39.3 Minutes since creation
       get.minutes.since.cookie <- function(cookie_name)
       {
-      #Cookie path
-        cookies_dir <- paste0(get.groundhog.folder(),"/cookies")
-        cookie_path <- file.path(cookies_dir, paste0(cookie_name,".csv"))
-        
+      #paths
+          cookies_dir <- paste0(get.groundhog.folder(),"/cookies")
+          cookie_path <- file.path(cookies_dir, paste0(cookie_name,".csv"))
+
       #Exists? Return 999999 if it does not, contents if it does
-        if (!file.exists(cookie_path)) return (999999)
-        if (file.exists(cookie_path)) {
+        if (!file.exists(cookie_path)) return (99999999)
+        if (file.exists(cookie_path))  {
           time0 <- utils::read.csv(cookie_path)$x  
           seconds <- as.numeric(Sys.time()-time0)
           minutes <- seconds/60
           return(minutes)
         }
      }#End of read cookie
-   
+      
+  
+    #39.4 Save Session cookie
+      save.session.cookie<-function(cookie_name)
+      {
+        .pkgenv[[cookie_name]] <- Sys.time()
+        
+      }
+      
+    #39.5
+      get.minutes.since.session.cookie<-function(cookie_name)
+      {
+        #Not set, return 99999
+          if (is.null(.pkgenv[[cookie_name]])) return(99999999)
+      
+        #If set, 
+          mins <- difftime(Sys.time(), .pkgenv[[cookie_name]] , units = 'mins')
+          return(mins)
+      }
+        
+        
    
 #40 get.repos_from.snowball()
       
       get.repos_from.snowball <- function(snowball)
       {
       #If missing, or explicitly MRAN or CRAN, then it is a CRAN package.
-        repos <- ifelse(snowball$from %in% c(NA< 'CRAN','MRAN'),'CRAN',snowball$from)
+        repos <- ifelse(snowball$from %in% c(NA,'', 'CRAN','GRAN','MRAN'),'CRAN',snowball$from)
       return(repos)
       }
 
     
-#41 Update groundhog session with new packages loaded and requested with groundhog
-      update.groundhog.session <- function(snowball)
+#41 Add snowball so session data.frame
+    
+      add.session.snowballs <- function(snowball)
       {
-      #The fucntion has as argument snowball_with_repos, to make sure the snowball submitted includes $repos
+      #The function has as argument snowball_with_repos, to make sure the snowball submitted includes $repos
       # this variable is set in groundhog.library.single() and groundhog.library.single.remote() just before
       # calling on the function update.groundhog.session()
         
+      #Add sha if CRAN (this allows later knowing if a package has been loaded from remote)
+          if (!'sha'  %in% names(snowball)) snowball$sha=NA
+        
       #Add repos
             snowball$repos <- get.repos_from.snowball(snowball)  #function 40,  just above
             
-      #Get requested pkg from snowball
-            pkg <- snowball$pkg[nrow(snowball)]
-          
-      #Subset from snowball
-				    groundhog.session_df.k <-snowball[,c('pkg','vrs','pkg_vrs','repos')]
+      #Subset of columns from snowball
+				    session.snowballs.k      <-snowball[,c('pkg','vrs','pkg_vrs','repos')]
 					  
 			#Add time
-					  groundhog.session_df.k$time <- as.numeric(Sys.time())
+					  session.snowballs.k$time <- as.numeric(Sys.time())
 					  
-			#TRUE FALSE for whether package was explicitly requested
-					  groundhog.session_df.k $requested <- groundhog.session_df.k$pkg == pkg     
-					  
-			#ADD NEW ROWS
-					  .pkgenv[['groundhog.session_df']] <- 	rbind(.pkgenv[['groundhog.session_df']] , groundhog.session_df.k )
+			#Requested
+					  session.snowballs.k$requested <- FALSE                  #No package was requested
+					  session.snowballs.k$requested[nrow(snowball)] <- TRUE   #Except the last one
+					
+			#Add to session
+					  .pkgenv[['session.snowballs']] <- 	rbind(.pkgenv[['session.snowballs']] , session.snowballs.k )
       }
       
-        
+ 
+      
 #42  check.groundhog.version()  - if more than `min.days' days since last check, check if groundhog needs to be updated
       
       check.groundhog.version <- function(min.days=7)
       {
-      ######
-        last.check.days=9999  #give high value so that when unassigned it works fine
+      #Early return if the folder with groundhog has not been set yet 
+          main_folder <-  fw(paste0(path.expand("~"), "/R_groundhog")) #fw: function #50
+          if (!file.exists(main_folder)) return(invisible(""))
+        
+      #Start with high numbers  
+        last.check.days=9999  
         
       #Skip cookie time if min.days=0
           if (min.days>0)
           {
         
-      #How many days has it been
+      #How many days has it been since we checked
           last.check.minutes <- get.minutes.since.cookie('check_groundhog_version')
           last.check.days    <- (last.check.minutes/60)/24
+          save.cookie('check_groundhog_version')
         
       #If less than min.days, early return
-          if (last.check.days<min.days) return(invisible(TRUE))
-          } 
-          #END IF MIN.days>0
-        
+          if (last.check.days<min.days) return(invisible(""))
+      
+          } #End if min.days>0
         
       #If more than `min.days` days, check version on server
           if (last.check.days>=min.days)
@@ -760,7 +812,7 @@ get.available.mran.date <- function(date0, date1) {
                 )
               
               #If NULL then early return
-                  if (is.null(groundhog.version_cran)) return(invisible(FALSE))
+                  if (is.null(groundhog.version_cran)) return(invisible(""))
               
               #Get majmin
                 groundhog.version_using <- as.character(packageVersion("groundhog"))
@@ -768,52 +820,91 @@ get.available.mran.date <- function(date0, date1) {
                 gv.cran  <- as.numeric(strsplit(groundhog.version_cran, "\\.")[[1]])
                 gv.using.majmin <-  10000*gv.using[1] + gv.using[2]
                 gv.cran.majmin  <-  10000*gv.cran[1]  + gv.cran[2]
-    
-    
+
               #If server's is bigger, prompt to update
                 if (isTRUE(gv.cran.majmin > gv.using.majmin)) {
-                    message2()
-                    message1(
-                    "\n\n\n",
-                    "          OUTDATED GROUNDHOG\n",
-                    "            You are using version  '" , groundhog.version_using, "\n",
-                    "            The current version is '" , groundhog.version_cran, "'\n\n",
-                    "            You can read about the changes here: https://groundhogr.com/changelog\n\n",
-                    "Update by running: \ninstall.packages('groundhog')"
+                    msg<-paste0(
+                    "########################################################\n",
+                    "OUTDATED GROUNDHOG\n",
+                    "You are using version  '" , groundhog.version_using, "'\n",
+                    "The current version is '" , groundhog.version_cran, "'\n\n",
+                    "You can read about the changes here:\n",
+                    "https://groundhogr.com/changelog\n\n",
+                    "Update by running: `install.packages('groundhog')`\n",
+                    "########################################################\n"
                     )
+                    return(msg)
                     }  #End mismatch in version
+                
                   
-            } #ENd last check more than `min.days`  ago
+            } #End last check more than `min.days`  ago
       }#End of function 42
       
-      
-#49 Check consent has been given to save files locally
-      check.consent <- function() {
+#43 Get operating system
+  get.os <- function()
+  {
+    
+  #Default
+    os <- 'other'
+
+  #If it has been set by hand, read it
+    path <-  paste0(path.expand("~"), "/R_groundhog/options/os.txt") #see groundhog.options.R
+    if (file.exists(path)) {
+      os = scan(path,what='character',quiet=TRUE)
+      if (os!='other') return(os)
+    }
+    
+    #There was a plan to have a groundhog.option() function but was abandoned.
+    #This is a placeholder, if there were to be an issue with some OS not being recognized
+    #users could by hand add this file, write their os and groundhog would work
+    #but if this ends up affecting more users groundhog will be changed to have a more robust os detection.
+    
+  # use contrib.url() to rely on R's processing of sys.info() alternatives
+     set.default.mirror()
+    repos <- as.character(getOption("repos"))
+    bin.url <- utils::contrib.url(repos,type='binary')
+    if (regexpr('windows', bin.url)[[1]]>0) os<-'windows'
+    if (regexpr('macosx', bin.url)[[1]]>0)  os<-'mac'
+    if (regexpr('arm64', bin.url)[[1]]>0 &  os=='mac') os<-'mac_arm'
+  
+    
+  #If other show warning
+    return(os)
+  }
+    
+
+    
+#44 Check consent has been given to save files locally
+      check.consent <- function(ask=TRUE) {
         
-      #Folder with cookie with location of groundohg folder, its exsitence means consent
-         main_folder <-  fw(paste0(path.expand("~"), "/R_groundhog")) #fw: function #50
+      #Folder with cookie with location of groundhog folder, its existence means consent
+         main_folder <-  fw(paste0(path.expand("~"), "/R_groundhog/")) #fw: function #50
         
       #See if consent has been given by seeing if the folder exists
-        consent <- (file.exists(main_folder))
+        consent <- (dir.exists(main_folder))
         if (consent==TRUE) return(TRUE)
+        if (consent==FALSE & ask==FALSE) return(FALSE)
+
         
        #If no consent, ask 
-         if (consent == FALSE) {
+         if (consent == FALSE & ask==TRUE) {
             msg       = paste0("groundhog needs authorization to save files to '",main_folder,
                         "'\n", "Enter 'OK' to provide authorization, and 'NO' not to.")
             
             batch_msg =  paste0("\n\n*********IMPORTANT MESSAGE FROM GROUNDHOG*****************************************\n",
-                         "groundhog needs to save files to '",main_folder,"' in order to work.\n",
+                         "groundhog needs to save packages in a local directory in order to work.\n",
                          "Per CRAN policy, you need to actively authorize groundhog to do this.\n",
-                         "To do so, simply create that folder, by e.g., running in R:\n",
-                         "       dir.create('",main_folder,"') \n",
+                         "Please run `set.groundhog.folder(<path>)` for the path you wish to use.\n",
+                         "If you are unsure which path to use, the following path is a reasonable default:\n",
+                         "   '", main_folder , "'\n",
+                         "(This only needs to be done once on a given computer)\n",
                          "**********************************************************************************\n\n")
           #For batch files
-            if (interactive()==FALSE)
+            if (interactive()==FALSE )
             {
               message(batch_msg)
-              return(FALSE)
-              
+              #return(FALSE)
+              exit()
             }
             
             
@@ -827,13 +918,337 @@ get.available.mran.date <- function(date0, date1) {
               }
             if (tolower(answer)=="no") {
               message("You did not provide permission to save files locally, groundhog will not work until you do.")
-              return(FALSE)
+              exit()
               }
-         } #end 
+         } #End  if no consent and we are asking for it
         
         
       }#End of function
           
-#50 Turn \ into /
+#45 Turn \ into /
       fw <- function(x) gsub("\\\\", "/", x)
-      
\ No newline at end of file
+      
+      
+
+#46 Moved to its own function
+#47 DROPPED
+#48 Moved to its own function
+ 
+  
+#49 - get.parallel.time()  :  estimate parallel time of installation
+get.parallel.time<-function(times,cores)
+{
+  #Sort times
+    times <- sort(times,decreasing=TRUE)
+    
+  #initiates times with small but sortable values
+    y = seq(0,.01,length.out=cores) 
+
+  #In loop, assign the next one, to the lowest total so far
+    for (k in 1:length(times))
+    {
+    y[which.min(y)]=y[which.min(y)] + times[k]  
+    }
+  #The longest link is the estimated time
+    return(max(y))
+}
+
+
+#50 get.pkg_list
+  get.remote_df.from.path<-function(snowball_path)
+  {
+    #Split path
+      parts <- strsplit(snowball_path,"/")[[1]]
+      n <- length(parts)
+      remote_id <- parts[n-1]
+      remote_id <- gsub("_", "", remote_id)
+    
+    #Split usr/pkg
+      parts2 <- strsplit(parts[n],"_")[[1]]
+      usr <- parts2[1]
+      pkg <- parts2[2]
+      sha <- parts2[3]
+      
+    return(data.frame(remote_id = remote_id , usr=usr, pkg=pkg, sha=sha))  
+  }
+      
+      
+#51 stop
+  gstop <- function(msg,format=FALSE) {
+    #Format the message with line breaks and border if requested
+    if (format==TRUE) msg=format_msg(msg) #util #35
+    message1(msg)
+    message("**Groundhog stopped**")
+    exit()
+    }
+  
+#52 read/save purge and grounhog.installed. rds
+  
+  read.local.rds <- function(filename)
+  {
+    #Using R 
+    r.version    <- get.r.majmin()
+    
+    #Path to files
+      dir <- paste0(get.groundhog.folder() , "/R-" , r.version ,"/rds_files/")
+      path <- file.path(dir,filename)
+      
+      
+    #Read the file
+      df <- data.frame()
+      if (file.exists(path)) df<-readRDS(path)
+
+    #Return 
+      return(df)
+  }
+  
+  save.local.rds <- function(df, filename)
+  {
+    #Using R 
+    r.version    <- get.r.majmin()
+    
+    
+  
+    #Path to files
+      dir<- paste0(get.groundhog.folder() , "/R-" , r.version ,"/rds_files/")
+      path <- file.path(dir,filename)
+  
+     #Ensure dir exists  
+      dir.create(dir,showWarnings = FALSE,recursive = TRUE)
+  
+    #Save
+      saveRDS(df,path,version=2)
+  }
+  
+
+#53 Download in batches
+  download.files.in_batches <- function(url.files , zip.files , batch.size = 20)
+    {
+    #Sort alphabetically by package name
+      j  <- order(basename(url.files))
+      url.files <- url.files[j]
+      zip.files <- zip.files[j]
+    
+    #Split the vectors
+      #https://stackoverflow.com/questions/3318333/split-a-vector-into-chunks
+      url.split <- split(url.files, ceiling(seq_along(url.files)/batch.size))
+      zip.split <- split(zip.files, ceiling(seq_along(zip.files)/batch.size))
+      btot <- length(url.split)  
+      
+  
+
+    #Download in loop
+      for (bk in 1:btot)
+      {
+      if (btot>1) message2("Batch ",bk, " of ", btot,". Downloading the following files:")
+      message1("     ",paste(url.split[[bk]],collapse='\n     '))
+      
+      #In R-3.6.3 a warning is show about using a vector, but it works with vectors, so turning it off here
+        w0 <- getOption('warn')
+        options(warn=-1)      
+        
+        try (utils::download.file(url.split[[bk]], zip.split[[bk]],quiet=TRUE,method='libcurl'))
+        
+        #Turn warning back on
+        options(warn=w0) 
+          
+      
+      }
+      
+  }#End of #53
+  
+  
+#54 filesize_format
+  #(written by chatGPT, to avoid relying on utils:::)
+  filesize_format <- function(size_in_bytes) {
+  if(size_in_bytes < 1024) {
+    return(paste0(size_in_bytes, " B"))
+  }
+  units <- c("KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
+  u <- -1
+  while(size_in_bytes >= 1024 & u < length(units)) {
+    size_in_bytes <- size_in_bytes / 1024
+    u <- u + 1
+  }
+  return(paste0(round(size_in_bytes, 1), " ", units[u+1]))
+  }
+  
+  
+#55 get.restore.points  
+  
+  get.restore.points <- function()
+  {
+    #0 Check consent to avoid zzz..R triggering loop
+        if (check.consent(ask=FALSE)==FALSE) {  
+          return(data.frame(date='',days='')[0,])
+          }
+  
+    #1 Path
+      restore_dir <- paste0(get.groundhog.folder(),"/restore_points/", get.r.majmin())
+      dir.create(restore_dir, recursive = TRUE,showWarnings = FALSE)    
+
+      
+    #2 Files available
+      restore_files <- list.files(restore_dir)
+        
+    #3 If none, end
+        if (length(restore_files)==0) {  
+          return(data.frame(date='',days='')[0,])
+          }
+        
+    #4 Turn filenames to dates
+      restore_dates <- as.Date(substr(restore_files, 0,10))  
+    
+    #5 Make data.frame
+      df <- data.frame(dates=restore_dates, days = as.numeric(Sys.Date()- restore_dates))
+      
+    #6 Return it
+      return(df)
+    
+  }
+  
+#57 view conflicts
+  view.conflicts.function <- function() {
+    return(.pkgenv[['conflicts']])
+    }
+
+  
+#58 get.ip
+  get.ip <- function(location)
+  {
+    #1 Get all subfolders for backup and groundhog
+    
+        #1.1 For groundhog. backup, all_local there is a folder with subfolders for each pkg, get all pkgs
+            if (location %in% c('backup','groundhog'))
+                {
+                 #Path containing all subfolders with pkg_vrs   
+                  if (location=='groundhog') {
+                        cran_path   <- paste0(get.groundhog.folder() , "/R-" , get.r.majmin())
+                        github_path <- paste0(cran_path,"/_github")
+                        gitlab_path <- paste0(cran_path,"/_gitlab")
+                        master_path=c(cran_path, github_path, gitlab_path)
+                      }
+                    
+                    
+                  if (location=='backup')    master_path <- paste0(get.groundhog.folder(),"/restore_library/" , get.r.majmin() , "/")
+
+                #All pkgs in that path    
+                  all.paths<- list.files(master_path,full.names=TRUE)
+            }
+
+
+        #1.2 Local  
+          if (location=='local')     all.paths <- .pkgenv[["orig_lib_paths"]][1]
+          if (location=='all_local') all.paths <- .pkgenv[["orig_lib_paths"]][-length(.pkgenv[["orig_lib_paths"]])]
+
+
+     #2 Get the installed.packages
+        ip <- data.frame(utils::installed.packages(all.paths), row.names = NULL, stringsAsFactors = FALSE)
+         
+     #3 Create pkg_vrs
+        if (nrow(ip)>0)  ip$pkg_vrs <- paste0(ip$Package,"_",ip$Version)
+        if (nrow(ip)==0) ip$pkg_vrs <- character()
+        
+      
+     #4 Select columns
+        ip <- ip[,c(names(ip) %in% c("LibPath", "Package","Version","pkg_vrs"))]
+        
+     #5 Add MD5 for DESCRIPTION file to merge & compare with `loans`
+        if (nrow(ip)>0)  {
+            description.path <- paste0(ip$LibPath, "/" , ip$Package , "/DESCRIPTION")
+            ip$md5 <- tools::md5sum(description.path)
+            }
+          
+        if (nrow(ip)==0) ip$md5 <- character()
+        
+    #6 End
+        return(ip)
+    }
+        
+
+#60 Loans
+  #60.1 get 
+  #pkg_vrs, location, 
+  get.loans<-function(verfiy.package.exists=TRUE) {
+    
+    #Start empty
+    loans <- data.frame(pkg_vrs=character() , 
+                        groundhog_location=character(),    
+                        sha=character(),                           
+                        md5=character(),                           
+                        stringsAsFactors=FALSE)
+    #Read it if it exists
+      loans_path<-paste0(get.groundhog.folder(),"/loans/",get.r.majmin(),".rds")  
+      dir.create(dirname(loans_path), showWarnings = FALSE,recursive = TRUE)
+      if (file.exists(loans_path)) loans<-readRDS(loans_path)
+      
+    #Sort it
+      loans<-loans[order(loans$pkg_vrs),]
+      
+    #Verify package exists in local folder
+      if (verfiy.package.exists==TRUE)
+      {
+        ip.local <- get.ip('local') #utils #58
+        loans   <- loans[loans$md5 %in% ip.local$md5,]
+      }
+       
+    #output
+      return(loans)
+  }
+  
+  #60.2 Save loans
+   save.loans<-function(loans) {
+     loans_path<-paste0(get.groundhog.folder(),"/loans/",get.r.majmin(),".rds")  
+     dir.create(dirname(loans_path), showWarnings = FALSE,recursive = TRUE)
+     saveRDS(loans,loans_path,version=2,compress=FALSE)
+   }
+
+   
+ #61 Verify unzipping has finished
+   
+    
+     verify.unzip<-function(zipfile,outdir)
+   
+      {
+        #Extension
+          ext <- tools::file_ext(zipfile)
+
+        #pkg name
+          pkg <- get.pkg(basename(zipfile))
+        
+        #File listed inside the zip file
+          if (ext=='zip') files.in.zip    = utils::unzip(zipfile , list=T) #(does not really unzip, just lists contents)
+          if (ext=='tar') files.in.zip    = utils::untar(zipfile , list=T) 
+        
+        #Files already outzipped
+          files.out.zip   = list.files(outdir ,recursive = TRUE  ,all.files = TRUE,include.dirs = TRUE)
+
+        #If same number, verify is TRUE
+          verify <- FALSE
+          if (length(files.out.zip)==length(files.in.zip$Name)) verify <- TRUE
+
+           
+        return(verify)
+      
+     }  
+     
+   #62 get drive from path whether they use forward or backwards slashes
+   get.drive=function(path)
+    {
+    parts  <- strsplit(path,"/")[[1]]            
+    drive1 <- ifelse(parts[1]=='',parts[2],parts[1])  #if path='/home/' then it starts with "/" and we want instead what comes next
+    drive2 <- strsplit(drive1,"\\\\")[[1]][1]         #need to double escape to get \\
+    drive2 <- tolower(drive2)
+    return(drive2)
+  }
+
+
+
+
+  #63 Slower with dropbox
+   groundhog.in.dropbox<-function()
+   {
+    return(regexpr('dropbox', tolower(get.groundhog.folder()))>0)
+        
+   }
+   
\ No newline at end of file
